using System;
using System.Collections.Generic;
using Google.Protobuf;
using GooglePB = global::Google.Protobuf;

using com.unity.mgobe.src.EventUploader;
using com.unity.mgobe.src.Util;
using com.unity.mgobe.src.Util.Def;

namespace com.unity.mgobe.src.Net {
    public class NetClient : Net {

        private readonly int _maxDataLength = Convert.ToInt32 (Math.Pow (2, 12));
        private readonly BstCallbacks _bstCallbacks;
        private static Dictionary<int, Action<byte[]>> _requestMap;
        
        public NetClient (BstCallbacks bstCallbacks) {
            this._bstCallbacks = bstCallbacks;
        }

        // 发送消息请求
        public string SendRequest (Object body, int subcmd, NetResponseCallback response, Action<ResponseEvent> callback, string cmd, string seq) {
            if (seq.Length == 0) {
                seq = Guid.NewGuid ().ToString ();
                var sendQueueVal = new SendQueueValue {
                    Time = DateTime.Now,
                        IsSocketSend = false,
                        Cmd = (int) subcmd,
                        resend = () => this.SendRequest (body, subcmd, response, callback, cmd, seq),
                        response = msg => {
                            response (true, msg, callback);
                            DeleteSendQueue (seq);
                        }
                };
                sendQueueVal.sendSuccess = () => {
                    // if(Socket.Id == 1) Debugger.Log("handle send success {0}", seq);
                    sendQueueVal.IsSocketSend = true;
                };
                sendQueueVal.remove = () => {
                    DeleteSendQueue (seq);
                };
                sendQueueVal.sendFail = (errCode, errMsg) => {
                    var errMessage = "消息发送失败，" + errMsg + "[" + errCode + "]";
                    var rspWrap1 = new ClientSendServerRspWrap1 {
                        Seq = seq,
                        ErrCode = errCode,
                        ErrMsg = errMessage
                    };
                    response (false, new DecodeRspResult {
                        RspWrap1 = rspWrap1,
                    }, callback);
                    DeleteSendQueue (seq);
                };
                AddSendQueue (seq, sendQueueVal);
            }

            // PB request = new PB();

            var qAppRequest = new ClientSendServerReqWrap1 {
                Version = RequestHeader.Version,
                AppName = RequestHeader.AppName,
                ClientIp = RequestHeader.ClientIp,
                ServiceIp = RequestHeader.ServiceIp,
                Business = RequestHeader.Business,
                AuthKey = RequestHeader.AuthKey,
                AuthType = RequestHeader.AuthType,
                AuthIp = RequestHeader.AuthIp,
                GameId = RequestHeader.GameId,
                Uid = RequestHeader.Uid,
                PlayerId = RequestHeader.PlayerId,
                Cmd = cmd,
                Seq = seq
            };
            var accessReq = new ClientSendServerReqWrap2 ();
            accessReq.Cmd = (ClientSendServerReqWrap2Cmd) subcmd;
            var data = Pb.EncodeReq (qAppRequest, accessReq, (GooglePB::IMessage)body);

            if (data.Length > _maxDataLength) {
                SendQueueValue val = null;
                SendQueue.TryGetValue(seq + "", out val);
                var timer = new Timer ();
                timer.SetTimeout (() => {
                    if (val != null) val.sendFail ((int) QAppProtoErrCode.EcSdkSendFail, "数据长度超限");
                }, 1);
                return seq;
            }

            var reqData = BuildData (data);

            return this.Send (reqData, seq, (ClientSendServerReqWrap2Cmd) subcmd);
        }

        private static byte[] BuildData (byte[] data) {
            return BuildData ((byte) MessageDataTag.ClientPre, data, (byte) MessageDataTag.ClientEnd);
        }

        // 接收响应并处理
        public void HandleMessage (byte[] body) {
            try {
                var rsp = Pb.DecodeRsp (body, (_seq) =>
                {
                    SendQueueValue _val = null;
                    SendQueue.TryGetValue(_seq + "", out _val);

                    if (_val == null)
                    {
                        return -1;
                    }
                    
                    return _val.Cmd;
                });
                
                var seq = rsp.RspWrap1.Seq;

                SendQueueValue val = null;
                SendQueue.TryGetValue(seq + "", out val);

                var callback = val?.response;

                if (val == null) return;
                // 处理错误码，并拦截 value.response

                EventUpload.PushRequestEvent (new ReqEventParam { RqCmd = val.Cmd, RqSq = rsp.RspWrap1.Seq, RqCd = rsp.RspWrap1.ErrCode, Time = Convert.ToInt64 ((DateTime.Now - val.Time).TotalMilliseconds) });

                // 心跳不拦截
                if (val.Cmd != (int) ClientSendServerReqWrap2Cmd.ECmdHeartBeatReq && HandleErrCode (rsp.RspWrap1)) {
                    return;
                }

                callback?.Invoke (rsp);
                return;
            } catch (Exception e) {
                Debugger.Log (e.ToString ());
            }
        }

        // 处理登录失败
        private void HandleTokenErr () {
            // 重登录
            UserStatus.SetStatus (UserStatus.StatusType.Logout);
            this.Socket.Emit ("autoAuth", null);
        }

        // 处理checklogin connect失败
        private void HandleRelayConnectErr () {
            Debugger.Log ("handle relay connect err");
            // 重checklogin
            CheckLoginStatus.SetStatus (CheckLoginStatus.StatusType.Offline);
            this.Socket.Emit ("autoAuth", null);

        }

        // 处理异常错误码
        // 返回 true 会拦截 responses 回调
        private bool HandleErrCode (ClientSendServerRspWrap1 res) {
            // Debugger.Log("handle errcode {0}", res.ErrCode);
            if (IsTokenError (res.ErrCode)) {
                this.HandleTokenErr ();
                Debugger.Log ("TOKEN_ERROR", res);
                return true;
            }

            if (IsRelayConnectError (res.ErrCode) && this.Socket.Id == (int) ConnectionType.Relay) {
                this.HandleRelayConnectErr ();
                Debugger.Log ("RELAY_CONNECT_ERROR", res);
                return true;
            }
            
            return false;
        }

        private static bool IsTokenError (int errCode) {
            var res = errCode == ErrCode.EcAccessCmdGetTokenErr ||
                errCode == ErrCode.EcAccessCmdTokenPreExpire ||
                errCode == ErrCode.EcAccessCmdInvalidToken ||
                errCode == ErrCode.EcAccessGetCommConnectErr;

            return res;
        }

        private static bool IsRelayConnectError (int errCode) {
            var res = errCode == ErrCode.EcAccessGetRelayConnectErr;
            return res;
        }

        // 如果返回码正确
        public static void HandleSuccess (int code, Action callback) {
            if (code == (int) QAppProtoErrCode.EcOk) {
                callback ();
            }
        }
    }
}